#!/usr/bin/env python3
#
# Module to load and expose file-location config
#
'''
Copyright (c) 2023 B Tasker

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import os
import re

CONFIG_BASE = os.getenv("CONFIG_BASE", ".")
SITE_LIST = []
SKIP_STRINGS = False
SKIP_REGEXES = False
SITE_PERMIT_RULES = False


def getConcurrency():
    ''' How many threads should multi-threaded operations use?
    '''
    return int(os.getenv("NUM_THREADS", 1))

def getPathHints():
    ''' Load a list of path hints from config
    '''
    hints = []
    hint_list = []
    # Load a list of URLs from the config file
    if os.path.exists(f"{CONFIG_BASE}/config/path-hints.txt"):
        with open(f"{CONFIG_BASE}/config/path-hints.txt", 'r') as f:
            hints = f.read().splitlines()
    
    for hint in hints:
        if hint.startswith("#"):
            continue
        hint_list.append(hint.strip())
    return hint_list

def getSites():
    
    # Load a list of URLs from the config file
    with open(f"{CONFIG_BASE}/config/sites.txt", 'r') as f:
        sites = f.read().splitlines()
    
    SITE_LIST = []
    for site in sites:
        site_l = site.strip().lower()
        if not site_l.startswith("http"):
            continue
        SITE_LIST.append(site_l.rstrip("/"))
    return SITE_LIST

def getExcludeStrings():
    l = []
    
    if os.path.exists(f"{CONFIG_BASE}/config/skipstrings.txt"):
        with open(f"{CONFIG_BASE}/config/skipstrings.txt") as f:
            l = f.read().splitlines()
    
    s = []
    for line in l:
        if len(line) > 2:
            s.append(line.split(" #")[0].strip())
    return s

def getExcludeRegexes():
    l = []
    
    if os.path.exists(f"{CONFIG_BASE}/config/skipregexes.txt"):
        with open(f"{CONFIG_BASE}/config/skipregexes.txt") as f:
            l = f.read().splitlines()
    
    s = []
    for line in l:
        if len(line) > 2:
            s.append(line.split(" #")[0].strip())
    return s


def getIgnoredTags():
    l = []
    
    if os.path.exists(f"{CONFIG_BASE}/config/ignoretags.txt"):
        with open(f"{CONFIG_BASE}/config/ignoretags.txt") as f:
            l = f.read().splitlines()
    
    s = []
    for line in l:
        if len(line) > 2:
            s.append(line.split(" #")[0].strip().lower())
    return s


def getImpliedTags():
    ''' Get a mapping of tags to imply if a specific tag exists
    
    For example, if we read a line
    
       foo!#!bar,sed 
       
    Then any content tag with foo should also be treated as if it's
    tagged with bar and sed 
    
    Introduced in utilities/file_location_listing#36
    '''
    l = []
    if os.path.exists(f"{CONFIG_BASE}/config/impliedtags.txt"):
        with open(f"{CONFIG_BASE}/config/impliedtags.txt") as f:
            l = f.read().splitlines()    

    s = {}
    for line in l:
        if len(line) > 2 and not line.startswith("#"):
            line_spl = line.lower().split("!#!")
            tag = line_spl[0].strip()
            implies = line_spl[1].strip()
            if tag not in s:
                # Split the implies entry into a list
                s[tag] = implies.split(",")
    return s


def getPermitRules():
    ''' Permit rules are a set of site specific regular expressions
    they can be used as an allowlist in order to constrain the 
    crawler to specific paths on that domain
    '''
    
    if os.path.exists(f"{CONFIG_BASE}/config/site-allowregexes.txt"):
        with open(f"{CONFIG_BASE}/config/site-allowregexes.txt") as f:
            l = f.read().splitlines()
    
    s = {}
    for line in l:
        if len(line) > 2 and not line.startswith("#"):
            line_spl = line.split("!#!")
            dom = line_spl[0].strip()
            rule = line_spl[1].strip()
            if dom not in s:
                # Pre-seed with a rule to match the very base of the domain
                # otherwise crawls just won't work
                s[dom] = [f'^http(s)?://{dom}$']
            
            s[dom].append(rule)               
    
    return s

def getDBPath():
    return os.getenv("DB_PATH", "/home/ben/tmp/search_db")

def getTagsEnabled():
    return (os.getenv("TAGS_ENABLED", "y").lower() == "y" )

def getUA():
    return os.getenv("UA", "location-lister")
    
def getForceReindex():
    return (os.getenv("REINDEX", "N") == "Y")

def getRevalidateCount():
    return int(os.getenv("REVAL_COUNT", 500))

def shouldSkipURL(url, domain):
    ''' Check an URL against excluded strings and regexes
    '''
    global SKIP_STRINGS
    global SKIP_REGEXES
    global SITE_PERMIT_RULES
    
    if not SITE_PERMIT_RULES:
        SITE_PERMIT_RULES = getPermitRules()
        
    # See whether there are any permit rules for this domain
    if domain in SITE_PERMIT_RULES:
        # There are, iterate through applying the regexes
        # to see whether any pass. If they do, the URL should
        # **not** be skipped
        for r in SITE_PERMIT_RULES[domain]:
            pattern = re.compile(r)
            if pattern.search(url):
                return False
        
        # Otherwise, none of the rules matched, skip 
        return True    
    

    if not SKIP_STRINGS:
        SKIP_STRINGS = getExcludeStrings()
        
    # Does the path contain a skipstring?
    for skipstr in SKIP_STRINGS:
        if skipstr in url:
            return True
    
    if not SKIP_REGEXES:
        SKIP_REGEXES = getExcludeRegexes()
        
    for r in SKIP_REGEXES:
        pattern = re.compile(r)
        if pattern.search(url):
            return True

    
