#!/usr/bin/env python3
#
# Storage module
#
'''
Copyright (c) 2023 B Tasker

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
import hashlib
import gzip
import json
import os
import sys
import tempfile
import time

from functools import lru_cache
from pathlib import Path
from subprocess import Popen
from urllib.parse import urlparse

sys.path.append(os.path.join(os.path.dirname(__file__), '../lib'))

import config


INDEXES = False # This will be populated on first index load
DB_PATH = '/opt/search_db'
INDEX_ALL = False
LOADED_AT = False
CONCURRENCY = config.getConcurrency()


def loadIndex():
    ''' Load the index from disk
    
    '''
    
    global LOADED_AT
    global INDEXES
    
    # Do we already have an index loaded in memory?
    if not INDEXES or len(INDEXES["IMAGE"]) > 0 or len(INDEXES["ALL"]) > 0 or len(INDEXES["DOC"]) > 0:
        # Blank it
        INDEXES = {
            "ALL" : [],
            "IMAGE" : [],
            "DOC" : [],
            "TAGS" : {},
            "IMPLIED" : []
        }
    start = time.time_ns()
    
    # Load details of any implied tags
    implied_tags = config.getImpliedTags()
    
    # read the tag index first
    url_tags = {}
    tag_id = 0
    if config.getTagsEnabled() and os.path.exists(f"{DB_PATH}/tags"):
        with gzip.open(f"{DB_PATH}/tags", 'r') as f:
            for l in f.read().splitlines():
                # Split the line up
                l_sp = l.decode().split("!#!")
                
                # Extract the tag
                k = l_sp[0]
                
                # Put the end of the line backtogether
                # even in the unlikely event our seperator
                # exists in it
                try:
                    v = json.loads("!#!".join(l_sp[1:]))
                except Exception as e:
                    print(f"Unable to load keys for tag {k}")
                    continue
                
                # Add the tag and it's calculated ID to the
                # tag list
                # check if it's there to prevent dupes
                if k not in INDEXES['TAGS']:
                    INDEXES['TAGS'][k] = tag_id
                    mytag = tag_id
                else:
                    mytag = INDEXES['TAGS'][k]
                
                # Update each of the keys to note it has this tag
                for url in v:
                    if url not in url_tags:
                        url_tags[url] = []
                
                    url_tags[url].append(tag_id)
                
                # Check whether this tag leads to any implied tags
                if k in implied_tags:
                    for implied_tag in implied_tags[k]:
                        # increment the tag id
                        tag_id += 1
                        
                        # insert the tag 
                        INDEXES['TAGS'][implied_tag] = tag_id
                        
                        # Mark the ID as being for an implied tag
                        INDEXES["IMPLIED"].append(tag_id)
                        
                        # mark each fo the urls as being linked to this tag
                        _ = [url_tags[url].append(tag_id) for url in v]
                            
                
                # Increment the tag id
                tag_id += 1
    
    
    tombstones = {}
    if os.path.exists(f"{DB_PATH}/index.tomb"):
        with open(f"{DB_PATH}/index.tomb", 'r') as f:
            for l in f.read().splitlines():
                tombstones[l.strip()] = 1
    
    
    seen = {}
    dupes = 0
    
    with gzip.open(f"{DB_PATH}/index", 'r') as f:
          # We mark each entry with an index number
          # this allows the search to parallelise by having
          # threads only consider entries with their corresponding
          # index number
          idx_num = 0
          
          # Iterate through the loaded file
          for l in f.read().splitlines():
            index_l = l.decode().split(",!,")
            
            # Has the URL been tombstoned?
            if index_l[0] in tombstones:
                continue
            
            if index_l[0] in seen:
                dupes += 1
                print(f"Skipping duplicate key {index_l[0]}")
                continue

            
            # Extract the file path from the key
            linkparsed = urlparse(index_l[0])
            dom = linkparsed.netloc.lower()
            pth = linkparsed.path.lower()
            
            # Create the index item
            idxitem = {
                "u" : index_l[0], # url
                "h" : index_l[1], # index hash
                "t" : index_l[2], # content-type
                "n" : index_l[3], # title/name
                "i" : idx_num, # Index number
                "p" : pth, # path
                "d" : dom, # domain
                "k" : []
            }
            
            # See whether the key exists in url_tags
            # if so, we've some tag keywords to record against this
            # entry
            if idxitem['u'] in url_tags:
                idxitem['k'] = url_tags[idxitem['u']]
            
            # Append into the relevant index(es)
            
            # Should we be populating the "ALL" index
            #
            # This is off by default - generally speaking we
            # tend to want to search by type, so having ALL is
            # just wasting RAM
            if INDEX_ALL:
                INDEXES["ALL"].append(idxitem)
                
            # Put into either image or doc
            #
            # TODO: do we perhaps want a video index?
            if index_l[2] in ["image/jpeg", "image/png", "image/gif"]:
                INDEXES["IMAGE"].append(idxitem)
            else:
                INDEXES["DOC"].append(idxitem)
             
            # Mark the key as seen
            seen[index_l[0]] = 1
            
            idx_num += 1
            if idx_num >= CONCURRENCY:
                idx_num = 0

    LOADED_AT = time.time()
    INDEXES["last_mod"] = getIndexLastMod()
    stop = time.time_ns()
    loaded_items = len(seen.keys())
    print(f"Index loaded {loaded_items} keys in {stop-start}ns")
    if dupes > 0:
        print(f"WARN: {dupes} duplicate index items detected")
    
    # Flush the file-cache
    readStoredFile.cache_clear()


def deleteStoredItem(url):
    '''
    Return stored item for the url
    '''
    
    # calculate a hash of the URL
    fname = getUrlHash(url)
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    
    file_path = f"{dir_path}/{fname}"
    
    if not os.path.exists(file_path):
        return False
    
    #print(f"Will delete{file_path}")
    os.remove(file_path)
    
    # Check if that's resulted in an empty directory
    if len(os.listdir(dir_path)) == 0:
        try:
            os.rmdir(dir_path)
        except:
            # oh well, no harm done
            pass

    # Also need to check whether there's a thumbnail
    thumbs_path = getThumbsPath(fname)
    path_prefix = f"{thumbs_path}/{fname}"
    for ext in [".gif", ".jpg"]:
        file_path = f"{path_prefix}{ext}"
        if not os.path.exists(file_path):
            continue
        
        # Otherwise, remove it
        os.remove(file_path)
        
    # Tombstone the URL so it'll be ignored in the index
    # at next index reload
    with open(f"{DB_PATH}/index.tomb", 'a+') as f:
        f.write(f"{url}\n")
        
    
def getIndexLastMod():
    ''' Get the mtime of the index
    If there's a tombstone **and** it's more recent, that
    time is returned instead
    '''
    
    tomb = 0
    tags = 0
    
    if os.path.exists(f"{DB_PATH}/index.tomb"):
        tomb = os.path.getmtime(f"{DB_PATH}/index.tomb")
        
    if os.path.exists(f"{DB_PATH}/tags"):
        tags = os.path.getmtime(f"{DB_PATH}/tags")
        
    idx = os.path.getmtime(f"{DB_PATH}/index")
    
    return max([idx, tomb, tags])
    

def getIsIndexFresh():
    ''' Check whether the index has changed since it was loaded
    
    This is achieved by getting mtime and comparing to the mtime
    stored with the index
    '''
    return (getIndexLastMod() == INDEXES['last_mod'])

def getStoredInfo(url):
    # Return stored info for the url
    # TODO
    
    # calculate a hash of the URL
    fname = getUrlHash(url)
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    file_path = f"{dir_path}/{fname}"
    
    j = readStoredFile(file_path)
    if not j:
        return 'FFFFFF', 'FFFFFF', 'FFFFFF'
    
    # return
    # last_mod, content_hash, valid_date
    # Support for content hashing isn't implemented yet
    return j['last-mod'], 'FFFFFFFFF', j['valid-at']


def getStoredItem(url, fname=False):
    '''
    Return stored item for the url
    
    If a fname is provided, that's used instead
    of calculating from the URL
    '''
    
    if not fname:
        # calculate a hash of the URL
        fname = getUrlHash(url)
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    file_path = f"{dir_path}/{fname}"

    # read the file and return the JSON
    return readStoredFile(file_path)


@lru_cache(maxsize=10000)
def readStoredFile(file_path):
    ''' Open a storage file for reading and return the json
    
    Attempts to read the file using gzip (the new storage format)
    and falls back to plain (the old) if that fails
    
    The fallback will eventually be removed.
    '''
    
    if not os.path.exists(file_path):
        return False
    
    with gzip.open(file_path) as f:
        for line in f:
            pass
        
        # line is now the last line
        try:
            j = json.loads(line)
        except:
            print("Failed to load file")
            invalidateFile(file_path)
            return False    
        
    return j   


def getStoragePath(fname, full=True):
    ''' Calculate the directory path
    
    '''
    parent_dir1 = fname[0:2]
    parent_dir2 = fname[2:4]
    if full:
        return f"{DB_PATH}/store/{parent_dir1}/{parent_dir2}"
    else:
        return f"{parent_dir1}/{parent_dir2}"


def getThumbsPath(fname, full=True):
    ''' Calculate the directory path for a thumbnail
    
    '''
    parent_dir1 = fname[0:2]
    parent_dir2 = fname[2:4]
    if full:
        return f"{DB_PATH}/thumbs/{parent_dir1}/{parent_dir2}"
    else:
        return f"{parent_dir1}/{parent_dir2}"   
    
@lru_cache(maxsize=10000)
def getUrlHash(url):
    ''' Generate a hash of an URL 
    This is used to navigate the cache on disk
    '''
    h = hashlib.sha256()
    h.update(url.encode('utf-8'))
    return h.hexdigest()

def invalidateFile(filepath):
    ''' Mark a file as being invalid
    '''
    print(f"Marking as invalid: {filepath}")
    try:
        os.replace(filepath, f"{filepath}.invalid")    
    except Exception as e:
        print(f"Unable to mark invalid: {e}")
    

def indexStorage():
    ''' After a run, we need to rebuild the index
    
    This is essentially just an ordered list of URLs
    '''
    index_list = []
    keyword_d = {}
    do_tags = config.getTagsEnabled()
    ignored_tags = config.getIgnoredTags()
    
    for subdir, dirs, files in os.walk(f"{DB_PATH}/store"):
        for file in files:
            # Skip temporary and invalid files
            if file.endswith(".invalid") or file.endswith(".tmp"):
                continue
            
            filepath = os.path.join(subdir, file)
            entry = [False, False, False, False]
            keywords = False
            # Open the file and extract the key and hash
            try:
                with gzip.open(filepath, 'r') as f:
                    for line in f:
                        line = line.decode()
                        if line.startswith("KEY"):
                            entry[0] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("HASH"):
                            entry[1] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("CONTENTTYPE"):
                            entry[2] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("TITLE"):
                            entry[3] = ":".join(line.split(":")[1:]).strip()
                        if do_tags and line.startswith("TAGS"):
                            keywords = ":".join(line.split(":")[1:]).strip()

                            
            except:
                invalidateFile(filepath)
                            
            if entry[2] is False:
                entry[2] = "!unk!"
                
            if entry[3] is False:
                entry[3] = "!unk!"
                
            if keywords:
                for k in json.loads(keywords):
                    k_l = k.lower()
                    if k_l in ignored_tags:
                        # Don't write this tag to the index
                        continue
                    if k_l not in keyword_d:
                        keyword_d[k_l] = []
                    keyword_d[k_l].append(entry[0])
                    
            # We should now have an entry with at least 2 parts. If we don't
            # skip this file
            if len(entry) < 4 or False in entry:
                continue
            
            # The index entry is:
            # url, path
            # Calculate the path
            pth = getStoragePath(entry[1], False)
            index_list.append(f"{entry[0]},!,{entry[1]},!,{entry[2]},!,{entry[3]}")
            
    # Sort the list
    index_list.sort()
    
    # Write to the index
    with gzip.open(f"{DB_PATH}/index.tmp", 'wt') as f:
        f.write('\n'.join(index_list))
    
    if do_tags:
        with gzip.open(f"{DB_PATH}/tags.tmp", 'wt') as f:
            for tag in keyword_d:
                f.write(f"{tag}!#!{json.dumps(keyword_d[tag])}\n")
        
    # Put the indexes into use 
    os.replace(f"{DB_PATH}/index.tmp", f"{DB_PATH}/index")
    if os.path.exists(f"{DB_PATH}/tags.tmp"):
        os.replace(f"{DB_PATH}/tags.tmp", f"{DB_PATH}/tags")    
    
    # Remove the tombstone file, if it exists
    if os.path.exists(f"{DB_PATH}/index.tomb"):
        os.remove(f"{DB_PATH}/index.tomb")
        


def storeInfo(pd):
    # Write out to storage 
    # TODO
    # For now, this just prints
    
    # Serialise the keywords
    kw_line = ""
    if len(pd['keywords']) > 0:
        kw = json.dumps(pd['keywords'])
        kw_line = f"TAGS:{kw}\n"
    
    file_str = (
        f"KEY:{pd['key']}\n"
        f"HASH:{pd['hashkey']}\n"
        f"DOMAIN:{pd['domain']}\n"
        f"PATH:{pd['path']}\n"
        f"FILE:{pd['filename']}\n"
        f"CONTENTTYPE:{pd['content-type']}\n"
        f"TITLE:{pd['title']}\n"
        f"{kw_line}"
        "\n"
    )
    file_str += json.dumps(pd)
    
    #print("===== ENTRY =====")
    #print(file_str)
    #print("===== ENTRY ENDS =====")
    
    # Store to disk
    dir_path = getStoragePath(pd['hashkey'])
    
    # ensure it exists
    path = Path(dir_path)
    path.mkdir(parents=True, exist_ok=True)
    
    # Open the file for writing
    fh = gzip.open(f"{dir_path}/{pd['hashkey']}.tmp", 'wt')
    fh.write(file_str)
    fh.close()
    
    # Move into place
    os.replace(f"{dir_path}/{pd['hashkey']}.tmp", f"{dir_path}/{pd['hashkey']}")
    
    
def storeImage(url, pg, contenttype):
    ''' We've fetched an image
    
    Thumbnail it
    '''
    fname = getUrlHash(url)
    
    thumbs_path = getThumbsPath(fname, full=True)
    
    # ensure it exists
    path = Path(thumbs_path)
    path.mkdir(parents=True, exist_ok=True)    
    
    # Save the main image into a temporary file
    tf = tempfile.NamedTemporaryFile(delete=True)
    tf.write(pg.content)
    tf.seek(0)
    
    if contenttype in ["image/gif"]:
        print(f"Thumbnailing gif {url}")
        # Can't thumnail a gif into a jpg, you just get lots of individual frames
        # Thumbnail it
        out, err = Popen(["convert",tf.name,"-loop", "0", "-resize","x450","-quality","75",f"{thumbs_path}/{fname}.tmp.gif"]).communicate()
        tf.close()   
        os.replace(f"{thumbs_path}/{fname}.tmp.gif", f"{thumbs_path}/{fname}.gif")
        
    else:
        # Thumbnail it
        out, err = Popen(["convert",tf.name,"-resize","x450","-quality","75",f"{thumbs_path}/{fname}.tmp.jpg"]).communicate()
        tf.close()   
        os.replace(f"{thumbs_path}/{fname}.tmp.jpg", f"{thumbs_path}/{fname}.jpg")
    
    return getThumbsPath(fname, full=False)
    

def buildPageDict(url, headers):
    ''' Take URL etc and create a dict using a standardised format
    
    This function is the authority on the structure of that format
    '''
    r = {
        "key" : url,
        "domain" : "",
        "path" : "",
        "filename" : "",
        "bytes" : False,
        "last-mod" : "FFFFFF",
        "content-type" : "FFFFFF",
        "title" : "",
        "valid-at" : time.strftime('%Y-%m-%d'),
        "image-path" : False,
        "keywords" : []
        }
    
    if headers.get("Last-Modified") is not None:
        r["last-mod"] = headers.get("Last-Modified")
        
    if headers.get("Content-Type") is not None:
        r["content-type"] = headers.get("content-type").lower()
        
    if headers.get("Content-Length") is not None:
        r["bytes"] = headers.get("Content-Length")
    
    r["hashkey"] = getUrlHash(url)
    
    linkparsed = urlparse(url)
    r["domain"] = linkparsed.netloc
    r["path"] = linkparsed.path
    r["filename"] = r["path"].split("/")[-1]
    
    return r

